from pymongo import MongoClient
from .. import local_config

class ConnectionMixin(object):
    def setUp(self):
        self.client = MongoClient(local_config.host, local_config.port)
        super(ConnectionMixin, self).setUp()



class QueryPlan(object):
    def __init__(self, db, collection_cls):
        self.db = db
        self.collection_cls = collection_cls
        self.collection_name = collection_cls.collection_name
        self.filters = {}
        self.ordering = None
        self.only_fields = None
        self._read_preference = None

    def _apply_aliasing(self, filters):
        if not hasattr(self.collection_cls, 'document_strategy'):
            return filters

        new_filters = {}
        for clause, expr in filters.iteritems():
            clause = self.collection_cls.document_strategy.get_alias(clause)
            new_filters[clause] = expr

        return new_filters

    def _cursor(self):
        coll = getattr(self.db, self.collection_name)

        filters = self._apply_aliasing(self.filters)

        kwargs = {}
        if self._read_preference:
            kwargs['read_preference'] = self._read_preference
        cursor = coll.find(filters, projection=self.only_fields, **kwargs)
        if self.ordering:
            cursor = cursor.sort(self.ordering)

        return cursor

    def filter(self, filters):
        self.filters = filters

    def only(self, fields):
        self.only_fields = fields

    def as_list(self):
        return [self._dict_to_object(doc) for doc in self._cursor()]

    def as_generator(self):
        for doc in self._cursor():
            yield self._dict_to_object(doc)

    def _dict_to_object(self, doc):
        return self.collection_cls.dict_to_object(doc, self.only_fields)

    def count(self):
        return self._cursor().count()

    def order_by(self, *ordering):
        self.ordering = list(ordering)

    def read_preference(self, read_preference):
        self._read_preference = read_preference



from .base import ConnectionMixin
from unittest import TestCase
from ..collection import Collection
from ..mapper import ClassFieldDelegator, SimpleMapper
from .. import fields
from .. import local_config


class Person(object):
    pass


class Athlete(Person):
    pass


class PeopleMapper(SimpleMapper):
    model = Person

    first_name = fields.StringField(field_name='f')
    last_name = fields.StringField(default='Smith')


class AthleteMapper(PeopleMapper):
    model = Athlete

    age = fields.IntField()


class ClassFieldMapper(ClassFieldDelegator):
    name_to_mapper = {
        'Athlete': AthleteMapper,
        'Person': PeopleMapper,
    }
    default_mapper = PeopleMapper


class People(Collection):
    collection_name = 'people'
    document_strategy = ClassFieldMapper


class TestPeristence(ConnectionMixin, TestCase):
    def setUp(self):
        super(TestPeristence, self).setUp()
        db_name = local_config.db_prefix + '_db'
        self.client.drop_database(db_name)
        self.db = getattr(self.client, db_name)

    def test_write_person(self):
        person = PeopleMapper.create(first_name='Colin', last_name='Howe')
        People.persist(self.db, person)

        query_plan = People.prepare_query(self.db)
        query_plan.filter({'first_name': 'Colin'})
        person, = query_plan.as_list()
        self.assertTrue(isinstance(person, Person))
        self.assertEquals(person.first_name, 'Colin')
        self.assertEquals(person.last_name, 'Howe')

    def test_write_athlete(self):
        person = AthleteMapper.create(
            first_name='Colin', last_name='Howe', age=28)
        People.persist(self.db, person)

        query_plan = People.prepare_query(self.db)
        query_plan.filter({'first_name': 'Colin'})
        person, = query_plan.as_list()
        self.assertTrue(isinstance(person, Athlete))
        self.assertEquals(person.first_name, 'Colin')
        self.assertEquals(person.last_name, 'Howe')
        self.assertEquals(person.age, 28)

    def test_update_person_post_create(self):
        person = AthleteMapper.create(
            first_name='Colin', last_name='Howe', age=28)
        People.persist(self.db, person)

        person.age = 29
        People.persist(self.db, person)

        query_plan = People.prepare_query(self.db)
        query_plan.filter({'first_name': 'Colin'})
        person, = query_plan.as_list()
        self.assertEquals(person.age, 29)

    def test_update_person_post_find(self):
        person = AthleteMapper.create(
            first_name='Colin', last_name='Howe', age=28)
        People.persist(self.db, person)

        query_plan = People.prepare_query(self.db)
        query_plan.filter({'first_name': 'Colin'})
        person, = query_plan.as_list()
        person.age = 29
        People.persist(self.db, person)

        person, = query_plan.as_list()
        self.assertEquals(person.age, 29)



from .base import ConnectionMixin
from unittest import TestCase
from ..collection import Collection
from ..mapper import ClassFieldDelegator, SimpleMapper
from .. import fields
from .. import local_config


class Person(object):
    pass


class Athlete(Person):
    pass


class SimplePeopleMapper(SimpleMapper):
    model = Person

    first_name = fields.StringField()
    last_name = fields.StringField(default='Smith')


class SimpleAthleteMapper(SimplePeopleMapper):
    model = Athlete


class ClassFieldMapper(ClassFieldDelegator):
    name_to_mapper = {
        'Athlete': SimpleAthleteMapper,
        'Person': SimplePeopleMapper,
    }
    default_mapper = SimplePeopleMapper


class People(Collection):
    collection_name = 'people'
    document_strategy = ClassFieldMapper


class TestSimpleMapper(ConnectionMixin, TestCase):
    def setUp(self):
        super(TestSimpleMapper, self).setUp()
        db_name = local_config.db_prefix + '_db'
        self.client.drop_database(db_name)
        self.db = getattr(self.client, db_name)

    def test_filter_returns_people(self):
        self.db.people.insert({'first_name': 'Colin'})

        query_plan = People.prepare_query(self.db)
        query_plan.filter({'first_name': 'Colin'})
        person, = query_plan.as_list()
        self.assertTrue(isinstance(person, Person))
        self.assertEquals(person.first_name, 'Colin')

    def test_default(self):
        self.db.people.insert({'first_name': 'Colin'})

        query_plan = People.prepare_query(self.db)
        person, = query_plan.as_list()
        self.assertEquals(person.last_name, 'Smith')

    def test_filter_returns_athletes(self):
        self.db.people.insert({'first_name': 'Colin', '_cls': 'Athlete'})

        query_plan = People.prepare_query(self.db)
        query_plan.filter({'first_name': 'Colin'})
        person, = query_plan.as_list()
        self.assertTrue(isinstance(person, Athlete))
        self.assertEquals(person.first_name, 'Colin')

    def test_subset_of_fields(self):
        self.db.people.insert({'first_name': 'Colin'})

        query_plan = People.prepare_query(self.db)
        query_plan.only('first_name')
        person, = query_plan.as_list()
        self.assertFalse(hasattr(person, 'last_name'))


class AliasedPeopleMapper(SimpleMapper):
    model = Person

    first_name = fields.StringField(field_name='f')
    last_name = fields.StringField(field_name='l')


class AliasedAthleteMapper(AliasedPeopleMapper):
    model = Athlete


class AliasedClassFieldMapper(ClassFieldDelegator):
    name_to_mapper = {
        'Athlete': AliasedAthleteMapper,
        'Person': AliasedPeopleMapper,
    }
    default_mapper = AliasedPeopleMapper


class Aliased(Collection):
    collection_name = 'aliased'
    document_strategy = AliasedClassFieldMapper


class TestAliasing(ConnectionMixin, TestCase):
    def setUp(self):
        super(TestAliasing, self).setUp()
        db_name = local_config.db_prefix + '_db'
        self.client.drop_database(db_name)
        self.db = getattr(self.client, db_name)

    def test_document_aliasing(self):
        self.db.aliased.insert({'f': 'Colin', 'l': 'Howe'})
        query_plan = Aliased.prepare_query(self.db)
        person, = query_plan.as_list()
        self.assertEquals(person.first_name, 'Colin')
        self.assertEquals(person.last_name, 'Howe')

    def test_query_aliasing(self):
        self.db.aliased.insert({'f': 'Colin', 'l': 'Howe'})
        self.db.aliased.insert({'f': 'Chris', 'l': 'Howe'})
        query_plan = Aliased.prepare_query(self.db)
        query_plan.filter({'first_name': 'Colin'})
        person, = query_plan.as_list()
        self.assertEquals(person.first_name, 'Colin')
        self.assertEquals(person.last_name, 'Howe')





from .base import ConnectionMixin
from unittest import TestCase
from ..collection import Collection
from ..mapper import SimpleMapper
from .. import fields
from .. import local_config


class Tag(object):
    pass


class TagMapper(SimpleMapper):
    model = Tag

    tag_id = fields.IntField(field_name='i')
    tag_name = fields.StringField(field_name='n')


class Person(object):
    pass


class PeopleMapper(SimpleMapper):
    model = Person

    first_name = fields.StringField(field_name='f')
    last_name = fields.StringField(default='Smith')
    tags = fields.ListField(mapper=TagMapper)


class People(Collection):
    collection_name = 'people'
    document_strategy = PeopleMapper


class TestListFields(ConnectionMixin, TestCase):
    def setUp(self):
        super(TestListFields, self).setUp()
        db_name = local_config.db_prefix + '_db'
        self.client.drop_database(db_name)
        self.db = getattr(self.client, db_name)

    def test_write_person(self):
        person = PeopleMapper.create(
            first_name='Colin', last_name='Howe',
            tags=[TagMapper.create(tag_id=1, tag_name='awesome')])
        People.persist(self.db, person)

        query_plan = People.prepare_query(self.db)
        query_plan.filter({'first_name': 'Colin'})
        person, = query_plan.as_list()
        self.assertTrue(isinstance(person, Person))
        self.assertEquals(person.first_name, 'Colin')
        self.assertEquals(person.last_name, 'Howe')
        self.assertTrue(isinstance(person.tags[0], Tag))

    def test_embedded_query(self):
        person = PeopleMapper.create(
            first_name='Colin', last_name='Howe',
            tags=[TagMapper.create(tag_id=1, tag_name='awesome')])
        People.persist(self.db, person)

        query_plan = People.prepare_query(self.db)
        query_plan.filter({'tags.tag_id': 1})
        person, = query_plan.as_list()
        self.assertTrue(isinstance(person, Person))
        self.assertEquals(person.first_name, 'Colin')
        self.assertEquals(person.last_name, 'Howe')

    def test_invalid_embedded_query(self):
        person = PeopleMapper.create(
            first_name='Colin', last_name='Howe',
            tags=[TagMapper.create(tag_id=1, tag_name='awesome')])
        People.persist(self.db, person)

        query_plan = People.prepare_query(self.db)
        query_plan.filter({'first_name.tag_id': 1})
        with self.assertRaises(ValueError):
            query_plan.as_list()



from .base import ConnectionMixin
from unittest import TestCase
from ..collection import Collection
from .. import local_config

class CustomPeopleStrategy(object):
    @staticmethod
    def dict_to_object(doc, only_fields):
        return Person(doc['first_name'])

    
    @classmethod
    def get_alias(cls, field_name):
        return field_name

class People(Collection):
    collection_name = 'people'
    document_strategy = CustomPeopleStrategy

class Person(object):
    def __init__(self, first_name):
        self.first_name = first_name

class TestCustomStrategy(ConnectionMixin, TestCase):
    def setUp(self):
        super(TestCustomStrategy, self).setUp()
        db_name = local_config.db_prefix + '_db'
        self.client.drop_database(db_name)
        self.db = getattr(self.client, db_name)

    def test_filter_returns_people(self):
        self.db.people.insert({'first_name': 'Colin'})

        query_plan = People.prepare_query(self.db)
        query_plan.filter({'first_name': 'Colin'})
        person, = query_plan.as_list()
        self.assertTrue(isinstance(person, Person))
        self.assertEquals(person.first_name, 'Colin')



from .base import ConnectionMixin
from unittest import TestCase
from ..collection import Collection
from .. import local_config

class People(Collection):
    collection_name = 'people'

class TestCollection(ConnectionMixin, TestCase):
    def setUp(self):
        super(TestCollection, self).setUp()
        db_name = local_config.db_prefix + '_db'
        self.client.drop_database(db_name)
        self.db = getattr(self.client, db_name)

    def test_unfiltered_count(self):
        self.db.people.insert({'first_name': 'Colin'})

        query_plan = People.prepare_query(self.db)
        self.assertEquals(1, query_plan.count())

    def test_filtered_count(self):
        self.db.people.insert({'first_name': 'Colin'})

        query_plan = People.prepare_query(self.db)
        query_plan.filter({'first_name': 'Colin'})
        self.assertEquals(1, query_plan.count())

        query_plan.filter({'first_name': 'Bob'})
        self.assertEquals(0, query_plan.count())



from fields import SimpleField, ComplexField

_field_cache = {}

def _populate_field_cache(cls):
    fields = {}
    for attr_name in dir(cls):
        attr = getattr(cls, attr_name)
        if isinstance(attr, SimpleField):
            fields[attr_name] = attr
    _field_cache[cls] = fields
            
def _get_field_cache(cls):
    fields = _field_cache.get(cls, {})
    if not fields:
        _populate_field_cache(cls)
        fields = _field_cache[cls]
    return fields

class SimpleMapper(object):
    @classmethod
    def dict_to_object(cls, doc, only_fields):
        fields = _get_field_cache(cls)
        model = cls.model()

        # Embedded documents do not always have IDs
        if '_id' in doc:
            model.id = doc['_id']

        for field_name, field_cls in fields.iteritems():
            if only_fields and field_name not in only_fields:
                continue
            db_field_name = field_cls.field_name or field_name
            if isinstance(field_cls, ComplexField):
                if db_field_name in doc:
                    field_value = field_cls.to_object(doc[db_field_name])
                else:
                    field_value = field_cls.default
            else:
                field_value = doc.get(db_field_name, field_cls.default)
            setattr(model, field_name, field_value)

        return model

    @classmethod
    def get_alias(cls, field_name):
        if '.' in field_name:
            field_name, children = field_name.split('.', 1)
        else:
            children = None

        fields = _get_field_cache(cls)
        if field_name in fields:
            field_cls = fields[field_name]

            if children:
                if not isinstance(field_cls, ComplexField):
                    raise ValueError('%s is not a ComplexField' % field_name)
                return "%s.%s" % (
                    field_cls.field_name or field_name,
                    field_cls.mapper.get_alias(children))
            else:
                return field_cls.field_name or field_name
        return None

    @classmethod
    def object_to_dict(cls, instance):
        fields = _get_field_cache(cls)

        dic = {}
        for field_name, field_cls in fields.iteritems():
            db_field_name = field_cls.field_name or field_name
            if isinstance(field_cls, ComplexField):
                value = field_cls.to_dict(instance, field_name)
            else:
                value = getattr(instance, field_name)

            dic[db_field_name] = value

        return dic

    @classmethod
    def create(cls, *args, **kwargs):
        instance = cls.model()
        fields = _get_field_cache(cls)
        for field_name in kwargs.iterkeys():
            if field_name not in fields:
                raise ValueError('%s does not have field %s' % (
                    cls.model, field_name))
            setattr(instance, field_name, kwargs[field_name])
        return instance
    
class ClassFieldDelegator(object):
    @classmethod
    def dict_to_object(cls, doc, only_fields):
        cls_name = doc.get('_cls')
        mapper = cls.name_to_mapper.get(cls_name)
        if not mapper:
            mapper = cls.default_mapper

        return mapper.dict_to_object(doc, only_fields)

    @classmethod
    def get_alias(cls, field_name):
        strategies = set(cls.name_to_mapper.values())
        strategies.add(cls.default_mapper)

        for strategy in strategies:
            alias = strategy.get_alias(field_name)
            if alias:
                return alias
        return field_name

    @classmethod
    def _get_most_specific_mapper(cls, instance):
        match_name, match_mapper = None, None
        for cls_name, mapper in cls.name_to_mapper.iteritems():
            if isinstance(instance, mapper.model):
                if match_name:
                    if issubclass(mapper.model, match_mapper.model):
                        match_name = cls_name
                        match_mapper = mapper
                else:
                    match_name = cls_name
                    match_mapper = mapper

        if not match_name:
            raise ValueError('Unexpected type %s' % instance.__class__)

        return match_name, match_mapper

    @classmethod
    def object_to_dict(cls, instance):
        cls_name, mapper = cls._get_most_specific_mapper(instance)

        dic = mapper.object_to_dict(instance)
        dic['_cls'] = cls_name
        return dic



class SimpleField(object):
    def __init__(self, field_name=None, default=None, required=False):
        self.default = default
        self.field_name = field_name
        self.required = required


class StringField(SimpleField):
    def __init__(self, max_length=None, choices=None, **kwargs):
        self.max_length = max_length
        self.choices = choices
        super(StringField, self).__init__(**kwargs)


class IntField(SimpleField):
    pass

class FloatField(SimpleField):
    pass

class BooleanField(SimpleField):
    pass

class DateTimeField(SimpleField):
    pass

class ObjectIdField(SimpleField):
    pass

class ComplexField(SimpleField):
    def __init__(self, embedded_field=None, mapper=None, *args, **kwargs):
        super(ComplexField, self).__init__(*args, **kwargs)
        self.mapper = mapper
        self.embedded_field = embedded_field

class ListField(ComplexField):
    def to_dict(self, instance, field_name):
        values = getattr(instance, field_name)
        if self.mapper:
            values = [self.mapper.object_to_dict(value) for value in values]

        return values

    def to_object(self, values):
        if self.mapper:
            values = [
                self.mapper.dict_to_object(value, None) for value in values]
        return values

class DictField(ComplexField):
    def to_dict(self, instance, field_name):
        values = getattr(instance, field_name)
        values = {
            k: self.mapper.object_to_dict(v) for k, v in values.iteritems()
        }
        return values



from queryplan import QueryPlan

class Collection(object):
    @classmethod
    def prepare_query(cls, db):
        return QueryPlan(db, cls)

    @classmethod
    def _get_collection(cls, db):
        return getattr(db, cls.collection_name)

    @classmethod
    def persist(cls, db, instance):
        coll = cls._get_collection(db)

        if hasattr(instance, 'id'):
            return cls.save_existing(coll, instance)
        else:
            return cls.save_new(coll, instance)

    @classmethod
    def prepare_object(cls, instance):
        return cls.document_strategy.object_to_dict(instance)

    @classmethod
    def save_existing(cls, coll, instance):
        dic = cls.prepare_object(instance)
        dic['_id'] = instance.id
        coll.save(dic)
        return True

    @classmethod
    def save_new(cls, coll, instance):
        dic = cls.prepare_object(instance)
        new_id = coll.save(dic)
        instance.id = new_id
        return True

    @classmethod
    def dict_to_object(cls, doc, only_fields):
        strategy = getattr(cls, 'document_strategy')
        if strategy:
            doc = strategy.dict_to_object(doc, only_fields)
        return doc

